/*
Copyright (C) 1996-1997 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

// screen.c -- master for refresh, status bar, console, chat, notify, etc

#include "quakedef.h"

float		scr_con_current;
float		scr_conlines;		// lines of console to display

cvar_t		gamma			= {"gamma",				"0"};
cvar_t		viewsize		= {"viewsize",			"0"};
cvar_t		scr_conspeed	= {"scr_conspeed",		"15"};
cvar_t		scr_centertime	= {"scr_centertime",	"2"};
cvar_t		scr_showpause	= {"showpause",			"1"};
cvar_t		scr_printspeed	= {"scr_printspeed",	"8"};
cvar_t		gl_triplebuffer = {"gl_triplebuffer",	"1",	true };

extern		cvar_t	crosshair;

qboolean	scr_initialized;		// ready to draw

int			clearconsole;
int			clearnotify;

viddef_t	vid;				// global video state

qboolean	scr_disabled_for_loading;
qboolean	scr_drawloading;
float		scr_disabled_time;

qboolean	block_drawing;

void SCR_ScreenShot_f (void);

/*
===============================================================================

CENTER PRINTING

===============================================================================
*/

char		scr_centerstring[1024];
float		scr_centertime_start;	// for slow victory printing
float		scr_centertime_off;
int			scr_center_lines;
int			scr_erase_lines;
int			scr_erase_center;

/*
===============
SCR_CenterPrint
===============
*/
void SCR_CenterPrint (char *str)
{
	strncpy (scr_centerstring, str, sizeof(scr_centerstring)-1);
	scr_centertime_off = scr_centertime.value;
	scr_centertime_start = cl.time;

// count the number of lines for centering
	scr_center_lines = 1;
	while (*str)
	{
		if (*str == '\n')
			scr_center_lines++;
		str++;
	}
}

void SCR_DrawCenterString (void)
{
	char	*start;
	int		l;
	int		j;
	int		x, y;
	int		remaining;

// the finale prints the characters one at a time
	if (cl.intermission)
		remaining = scr_printspeed.value * (cl.time - scr_centertime_start);
	else
		remaining = 9999;

	scr_erase_center = 0;
	start = scr_centerstring;

	if (scr_center_lines <= 4)
		y = vid.height*0.35;
	else
		y = 48;

	do	
	{
	// scan the width of the line
		for (l=0 ; l<40 ; l++)
			if (start[l] == '\n' || !start[l])
				break;
		x = (vid.width - l*16)/2;
		for (j=0 ; j<l ; j++, x+=16)
		{
			Draw_Character (x, y, start[j]);	
			
			if (!remaining--)
			{
				return;
			}
		}
			
		y += 12;

		while (*start && *start != '\n')
			start++;

		if (!*start)
			break;
		start++;		// skip the \n
	} while (1);
}

void SCR_CheckDrawCenterString (void)
{
	if (scr_center_lines > scr_erase_lines)
		scr_erase_lines = scr_center_lines;

	scr_centertime_off -= host_frametime;
	
	if (scr_centertime_off <= 0 && !cl.intermission)
		return;
	if (key_dest != key_game)
		return;

	SCR_DrawCenterString ();
}

//============================================================================

/*
==================
SCR_Init
==================
*/
void SCR_Init (void)
{
	Cvar_RegisterVariable (&gamma);
	Cvar_RegisterVariable (&viewsize);
	Cvar_RegisterVariable (&scr_conspeed);
	Cvar_RegisterVariable (&scr_showpause);
	Cvar_RegisterVariable (&scr_centertime);
	Cvar_RegisterVariable (&scr_printspeed);
	Cvar_RegisterVariable (&gl_triplebuffer);

//
// register our commands
//
	Cmd_AddCommand ("screenshot",SCR_ScreenShot_f);

	scr_initialized = true;
}

// Tomaz - FPS Counter Begin
/*
===========
SCR_DrawFPS
===========
*/

float	fpscounter;
char	temp[32];

void SCR_DrawFPS (void)
{
	static double	currtime;
	double			newtime;
	int				calc;

	if ((cl.time + 1) < fpscounter)
		fpscounter = cl.time + 0.1;

	newtime		= Sys_FloatTime ();
	calc		= (int) ((1.0 / (newtime - currtime)) + 0.5);
	currtime	= newtime;

	if (cl.time > fpscounter)
	{
		sprintf(temp, "%3i", calc);
		fpscounter = cl.time + 0.1;
	}

	Draw_String (vid.width - sizeof(temp) - 8, 0, temp, false);
}
// Tomaz - FPS Counter End

/*
=========
DrawPause
=========
*/
void SCR_DrawPause (void)
{
	int	x, y;

	if (!scr_showpause.value)		// turn off for screenshots
		return;

	if (!cl.paused)
		return;

	x = (vid.width  * 0.5f) - 36;
	y = (vid.height * 0.5f) - 8;

	Draw_String (x, y, "Paused", false);
}

//=============================================================================

/*
======================
SCR_SetUpToDrawConsole
======================
*/
void SCR_SetUpToDrawConsole (void)
{
	Con_CheckResize ();
	
	if (scr_drawloading)
		return;		// never a console with loading plaque

// decide on the height of the console
	con_forcedup = !cl.worldmodel || cls.signon != SIGNONS;

	if (con_forcedup)
	{
		scr_conlines = vid.height;		// full screen
		scr_con_current = scr_conlines;
	}
	else if (key_dest == key_console)
		scr_conlines = vid.height/2;	// half screen
	else
		scr_conlines = 0;				// none visible
	
	if (scr_conlines < scr_con_current)
	{
		scr_con_current -= scr_conspeed.value;
		if (scr_conlines > scr_con_current)
			scr_con_current = scr_conlines;

	}
	else if (scr_conlines > scr_con_current)
	{
		scr_con_current += scr_conspeed.value;
		if (scr_conlines < scr_con_current)
			scr_con_current = scr_conlines;
	}

	else
		con_notifylines = 0;
}

int	notify_amount;

/*
===============
SCR_DrawConsole
===============
*/
void SCR_DrawConsole (void)
{
	notify_amount = 0;

	if (scr_con_current)
	{
		Con_DrawConsole (scr_con_current, true);
		clearconsole = 0;
	}
	else
	{
		if (key_dest == key_game || key_dest == key_message)
			Con_DrawNotify ();	// only draw notify in game
	}
}

/* 
============================================================================== 
 
						SCREEN SHOTS 
 
============================================================================== 
*/ 

typedef struct _TargaHeader
{
	unsigned char 	id_length, colormap_type, image_type;
	unsigned short	colormap_index, colormap_length;
	unsigned char	colormap_size;
	unsigned short	x_origin, y_origin, width, height;
	unsigned char	pixel_size, attributes;
} TargaHeader;

// Tomaz - Redone ScreenShot & Movie Exporter Begin
/* 
================== 
SCR_ScreenShot_f
================== 
*/  
void SCR_ScreenShot_f (void) 
{
	byte		*buffer;
	char		tganame[80]; 
	char		checkname[MAX_OSPATH];
	int			i, c, temp;
// 
// find a file name to save it to 
// 
	sprintf (checkname, "%s/screenies", com_gamedir);
	Sys_mkdir (checkname);	

	strcpy(tganame,"SWD00.tga");

	for (i=0 ; i<=99 ; i++) 
	{
		tganame[3] = (i/10)%10 + '0';
		tganame[4] = (i   )%10 + '0'; 			
		sprintf (checkname, "%s/screenies/%s", com_gamedir, tganame);
		if (Sys_FileTime(checkname) == -1)
			break;	// file doesn't exist
	} 
	if (i==100) 
	{
		Con_Printf ("SCR_ScreenShot_f: Couldn't create a TGA file\n"); 
		return;
 	}
	
	buffer = malloc(vid.width*vid.height*3 + 18);
	memset (buffer, 0, 18);
	buffer[2] = 2;		// uncompressed type
	buffer[12] = vid.width&255;
	buffer[13] = vid.width>>8;
	buffer[14] = vid.height&255;
	buffer[15] = vid.height>>8;
	buffer[16] = 24;	// pixel size

	glReadPixels (0, 0, vid.width, vid.height, GL_RGB, GL_UNSIGNED_BYTE, buffer+18 ); 

	// swap rgb to bgr
	c = 18+vid.width*vid.height*3;
	for (i=18 ; i<c ; i+=3)
	{
		temp = buffer[i];
		buffer[i] = buffer[i+2];
		buffer[i+2] = temp;
	}

	COM_WriteFile (va ("screenies/%s", tganame), buffer, vid.width*vid.height*3 + 18 );

	free (buffer);
	Con_Printf ("Wrote %s\n", tganame);
} 
// Tomaz - Redone ScreenShot & Movie Exporter End

//=============================================================================

/*
======================
SCR_BeginLoadingPlaque
======================
*/
void SCR_BeginLoadingPlaque (void)
{
	S_StopAllSounds (true);

	if (cls.state != ca_connected)
		return;
	if (cls.signon != SIGNONS)
		return;
	
// redraw with no console and the loading plaque
	Con_ClearNotify ();
	scr_centertime_off = 0;
	scr_con_current = 0;

	scr_drawloading = true;
	SCR_UpdateScreen ();
	scr_drawloading = false;

	scr_disabled_for_loading = true;
	scr_disabled_time = realtime;
}

/*
====================
SCR_EndLoadingPlaque
====================
*/
void SCR_EndLoadingPlaque (void)
{
	scr_disabled_for_loading = false;
	Con_ClearNotify ();
}

//=============================================================================

char	*scr_notifystring;
qboolean	scr_drawdialog;

void SCR_DrawNotifyString (void)
{
	char	*start;
	int		l;
	int		j;
	int		x, y;

	start = scr_notifystring;

	y = vid.height*0.35;

	do	
	{
	// scan the width of the line
		for (l=0 ; l<40 ; l++)
			if (start[l] == '\n' || !start[l])
				break;
		x = (vid.width - l*16)/2;
		for (j=0 ; j<l ; j++, x+=16)
			Draw_Character (x, y, start[j]);	
			
		y += 12;

		while (*start && *start != '\n')
			start++;

		if (!*start)
			break;
		start++;		// skip the \n
	} while (1);
}

/*
================
SCR_ModalMessage
================
*/
int SCR_ModalMessage (char *text)
{
	if (cls.state == ca_dedicated)
		return true;

	scr_notifystring = text;
 
// draw a fresh screen
	scr_drawdialog = true;
	SCR_UpdateScreen ();
	scr_drawdialog = false;
	
	S_ClearBuffer ();		// so dma doesn't loop current sound

	do
	{
		key_count = -1;		// wait for a key down and up
		Sys_SendKeyEvents ();
	} while (key_lastpress != 'y' && key_lastpress != 'n' && key_lastpress != K_ESCAPE);

	SCR_UpdateScreen ();

	return key_lastpress == 'y';
}

//=============================================================================

/*
====================
SCR_BringDownConsole
====================
*/
void SCR_BringDownConsole (void)
{
	int		i;
	
	scr_centertime_off = 0;
	
	for (i=0 ; i<20 && scr_conlines != scr_con_current ; i++)
		SCR_UpdateScreen ();

	cl.cshift.percent = 0;		// no area contents palette on next frame
	VID_SetPalette (host_basepal);
}

/*
================
SCR_UpdateScreen
================
*/

void SCR_UpdateScreen (void)
{
	if (block_drawing)
		return;

	if (scr_disabled_for_loading)
	{
		if (realtime - scr_disabled_time > 60)
		{
			scr_disabled_for_loading = false;
			Con_Printf ("load failed.\n");
		}
		else
			return;
	}

	if (!scr_initialized || !con_initialized)
		return;				// not initialized yet
	
	//
	// determine size of refresh window
	//

//
// do 3D refresh drawing, and then update the screen
//
	SCR_SetUpToDrawConsole ();
	
	V_RenderView ();
		
	GL_Set2D ();

	//muff - draw map shot until map is loaded and ready
	if (con_forcedup && sv.active)
	{ 
		Draw_MapShots ();
	}

	else if (scr_drawdialog)
	{
		Sbar_Draw ();
		Draw_FadeScreen ();
		SCR_DrawNotifyString ();
	}
	else if (scr_drawloading)
	{
		Draw_MapShots ();
	}
	else if (cl.intermission == 1 && key_dest == key_game)
	{
		Sbar_DeathmatchOverlay ();
	}
	else if (cl.intermission == 2 && key_dest == key_game)
	{
		SCR_CheckDrawCenterString ();
	}
	else
	{
		// Tomaz - Crosshair Begin
		if (crosshair.value)
		{	
			Draw_Crosshair();
		}	
		// Tomaz - Crosshair End

		SCR_DrawPause ();
		SCR_CheckDrawCenterString ();
		SCR_DrawConsole ();
		Sbar_Draw ();
		M_Draw ();
	}

	if (show_fps.value)
		SCR_DrawFPS ();	// Tomaz - FPS Counter

	V_UpdatePalette ();

	GL_EndRendering ();
}