/*
Copyright (C) 1996-1997 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/
// gl_warp.c -- sky and water polygons

#include "quakedef.h"

extern	model_t	*loadmodel;

msurface_t	*warpface;

extern cvar_t gl_subdivide_size;

void BoundPoly (int numverts, float *verts, vec3_t mins, vec3_t maxs)
{
	int		i, j;
	float	*v;

	mins[0] = mins[1] = mins[2] =  9999;
	maxs[0] = maxs[1] = maxs[2] = -9999;
	v = verts;
	for (i=0 ; i<numverts ; i++)
	{
		for (j=0 ; j<3 ; j++, v++)
		{
			if (*v < mins[j])
				mins[j] = *v;
			if (*v > maxs[j])
				maxs[j] = *v;
		}
	}
}

void SubdividePolygon (int numverts, float *verts)
{
	int		i, j, k;
	vec3_t	mins, maxs;
	float	m;
	float	*v;
	vec3_t	front[64], back[64];
	int		f, b;
	float	dist[64];
	float	frac;
	glpoly_t	*poly;
	float	s, t;

	if (numverts > 60)
		Sys_Error ("numverts = %i", numverts);

	BoundPoly (numverts, verts, mins, maxs);

	for (i=0 ; i<3 ; i++)
	{
		m = (mins[i] + maxs[i]) * 0.5;
		m = gl_subdivide_size.value * floor (m/gl_subdivide_size.value + 0.5);
		if (maxs[i] - m < 8)
			continue;
		if (m - mins[i] < 8)
			continue;

		// cut it
		v = verts + i;
		for (j=0 ; j<numverts ; j++, v+= 3)
			dist[j] = *v - m;

		// wrap cases
		dist[j] = dist[0];
		v-=i;
		VectorCopy (verts, v);

		f = b = 0;
		v = verts;
		for (j=0 ; j<numverts ; j++, v+= 3)
		{
			if (dist[j] >= 0)
			{
				VectorCopy (v, front[f]);
				f++;
			}
			if (dist[j] <= 0)
			{
				VectorCopy (v, back[b]);
				b++;
			}
			if (dist[j] == 0 || dist[j+1] == 0)
				continue;
			if ( (dist[j] > 0) != (dist[j+1] > 0) )
			{
				// clip point
				frac = dist[j] / (dist[j] - dist[j+1]);
				for (k=0 ; k<3 ; k++)
					front[f][k] = back[b][k] = v[k] + frac*(v[3+k] - v[k]);
				f++;
				b++;
			}
		}

		SubdividePolygon (f, front[0]);
		SubdividePolygon (b, back[0]);
		return;
	}

	poly = Hunk_Alloc (sizeof(glpoly_t) + (numverts-4) * VERTEXSIZE*sizeof(float));
	poly->next = warpface->polys;
	warpface->polys = poly;
	poly->numverts = numverts;
	for (i=0 ; i<numverts ; i++, verts+= 3)
	{
		VectorCopy (verts, poly->verts[i]);
		s = DotProduct (verts, warpface->texinfo->vecs[0]);
		t = DotProduct (verts, warpface->texinfo->vecs[1]);
		poly->verts[i][3] = s;
		poly->verts[i][4] = t;
	}
}

/*
===================
GL_SubdivideSurface
===================
*/
void GL_SubdivideSurface (msurface_t *s)
{
	vec3_t		verts[64];
	int			numverts;
	int			i;
	int			lindex;
	float		*vec;

	warpface = s;

	//
	// convert edges back to a normal polygon
	//
	numverts = 0;
	for (i=0 ; i<s->numedges ; i++)
	{
		lindex = loadmodel->surfedges[s->firstedge + i];

		if (lindex > 0)
			vec = loadmodel->vertexes[loadmodel->edges[lindex].v[0]].position;
		else
			vec = loadmodel->vertexes[loadmodel->edges[-lindex].v[1]].position;
		VectorCopy (vec, verts[numverts]);
		numverts++;
	}

	SubdividePolygon (numverts, verts[0]);
}

//=========================================================

// speed up sin calculations - Ed
float turbsin[] =
{
	#include "gl_warp_sin.h"
};

#define TURBSCALE (40.7436589469704879)	// Tomaz Speed

/*
==============
EmitWaterPolys
==============
*/
void EmitWaterPolys (msurface_t *s)
{
	glpoly_t	*p;
	float		*v;
	int			i;
	float		r, t, os, ot;

	for (p=s->polys ; p ; p=p->next)
	{
		glBegin (GL_POLYGON);
		for (i=0,v=p->verts[0] ; i<p->numverts ; i++, v+=VERTEXSIZE)
		{
			os = v[3];
			ot = v[4];

			r = os + turbsin[(int)((ot*0.125+realtime) * TURBSCALE) & 255];
			r *= (0.015f);	// Tomaz Speed

			t = ot + turbsin[(int)((os*0.125+realtime) * TURBSCALE) & 255];
			t *= (0.015f);	// Tomaz Speed
			
			glTexCoord2f (r, t);
			glVertex3fv (v);
		}
		glEnd ();
	}
}

// Tomaz - Skybox Start

qboolean	skybox_init = false;
char		skyname[256];
char		*suf[6] = {"ft", "bk", "rt", "lf", "up", "dn"};
int			skytexture[6];

/*
==========
R_LoadSkys
==========
*/
void R_LoadSkys (void)
{
	int		i;
	char	name[64];

	for (i=0 ; i<6 ; i++)
	{
		sprintf (name, "gfx/env/%s%s", skyname, suf[i]);
		skytexture[i] = loadtextureimage(name, false ,true);

		if (skytexture[i] == 0)
		{
			skybox_init	= false;
			return;
		}
	}
	skybox_init	= true;
}

void R_SetSkyBox (char *sky)
{
	strcpy(skyname, sky);
	R_LoadSkys ();
}

void LoadSky_f (void)
{
	switch (Cmd_Argc())
	{
	case 1:
		if (skyname[0])
			Con_Printf("Current sky: %s\n", skyname);
		else
			Con_Printf("Error: No skybox has been set\n");
		break;
	case 2:
		R_SetSkyBox(Cmd_Argv(1));
		break;
	default:
		if (!cls.demoplayback)
			Con_Printf("Usage: loadsky skyname\n");
		break;
	}
}
// Tomaz - Skybox End

/*
============
R_DrawSkyBox
============
*/

#define R_SkyBoxPolyVec(s,t,x,y,z) \
	glTexCoord2f(s, t);\
	glVertex3f((x) + r_refdef.origin[0], (y) + r_refdef.origin[1], (z) + r_refdef.origin[2]);

void R_DrawSkyBox (void)
{
	if (!skybox_init)
		return;

	glBindTexture(GL_TEXTURE_2D, skytexture[0]); // front
	glBegin(GL_QUADS);
		R_SkyBoxPolyVec (0.998f, 0.002f,  3072, -3072,  3072);
		R_SkyBoxPolyVec (0.998f, 0.998f,  3072, -3072, -3072);
		R_SkyBoxPolyVec (0.002f, 0.998f,  3072,  3072, -3072);
		R_SkyBoxPolyVec (0.002f, 0.002f,  3072,  3072,  3072);
	glEnd();

	glBindTexture(GL_TEXTURE_2D, skytexture[1]); // back
	glBegin(GL_QUADS);
		R_SkyBoxPolyVec (0.998f, 0.002f, -3072,  3072,  3072);
		R_SkyBoxPolyVec (0.998f, 0.998f, -3072,  3072, -3072);
		R_SkyBoxPolyVec (0.002f, 0.998f, -3072, -3072, -3072);
		R_SkyBoxPolyVec (0.002f, 0.002f, -3072, -3072,  3072);
	glEnd();

	glBindTexture(GL_TEXTURE_2D, skytexture[2]); // right
	glBegin(GL_QUADS);
		R_SkyBoxPolyVec (0.998f, 0.002f,  3072,  3072,  3072);
		R_SkyBoxPolyVec (0.998f, 0.998f,  3072,  3072, -3072);
		R_SkyBoxPolyVec (0.002f, 0.998f, -3072,  3072, -3072);
		R_SkyBoxPolyVec (0.002f, 0.002f, -3072,  3072,  3072);
	glEnd();

	glBindTexture(GL_TEXTURE_2D, skytexture[3]); // left
	glBegin(GL_QUADS);
		R_SkyBoxPolyVec (0.998f, 0.002f, -3072, -3072,  3072);
		R_SkyBoxPolyVec (0.998f, 0.998f, -3072, -3072, -3072);
		R_SkyBoxPolyVec (0.002f, 0.998f,  3072, -3072, -3072);
		R_SkyBoxPolyVec (0.002f, 0.002f,  3072, -3072,  3072);
	glEnd();

	glBindTexture(GL_TEXTURE_2D, skytexture[4]); // up
	glBegin(GL_QUADS);
		R_SkyBoxPolyVec (0.998f, 0.002f,  3072, -3072,  3072);
		R_SkyBoxPolyVec (0.998f, 0.998f,  3072,  3072,  3072);
		R_SkyBoxPolyVec (0.002f, 0.998f, -3072,  3072,  3072);
		R_SkyBoxPolyVec (0.002f, 0.002f, -3072, -3072,  3072);
	glEnd();

	glBindTexture(GL_TEXTURE_2D, skytexture[5]); // down
	glBegin(GL_QUADS);
		R_SkyBoxPolyVec (0.998f, 0.002f,  3072,  3072, -3072);
		R_SkyBoxPolyVec (0.998f, 0.998f,  3072, -3072, -3072);
		R_SkyBoxPolyVec (0.002f, 0.998f, -3072, -3072, -3072);
		R_SkyBoxPolyVec (0.002f, 0.002f, -3072,  3072, -3072);
	glEnd();
}

#define R_FoggedSkyBoxPolyVec(x,y,z) \
	glVertex3f((x) + r_refdef.origin[0], (y) + r_refdef.origin[1], (z) + r_refdef.origin[2]);

void R_DrawFoggedSkyBox (void)
{
	glDisable (GL_TEXTURE_2D);
	glColor3f (gl_fogred.value, gl_foggreen.value, gl_fogblue.value);

	glBegin(GL_QUADS);
		R_FoggedSkyBoxPolyVec ( 3072, -3072,  3072);
		R_FoggedSkyBoxPolyVec ( 3072, -3072, -3072);
		R_FoggedSkyBoxPolyVec ( 3072,  3072, -3072);
		R_FoggedSkyBoxPolyVec ( 3072,  3072,  3072);
	glEnd();

	glBegin(GL_QUADS);
		R_FoggedSkyBoxPolyVec (-3072,  3072,  3072);
		R_FoggedSkyBoxPolyVec (-3072,  3072, -3072);
		R_FoggedSkyBoxPolyVec (-3072, -3072, -3072);
		R_FoggedSkyBoxPolyVec (-3072, -3072,  3072);
	glEnd();

	glBegin(GL_QUADS);
		R_FoggedSkyBoxPolyVec ( 3072,  3072,  3072);
		R_FoggedSkyBoxPolyVec ( 3072,  3072, -3072);
		R_FoggedSkyBoxPolyVec (-3072,  3072, -3072);
		R_FoggedSkyBoxPolyVec (-3072,  3072,  3072);
	glEnd();

	glBegin(GL_QUADS);
		R_FoggedSkyBoxPolyVec (-3072, -3072,  3072);
		R_FoggedSkyBoxPolyVec (-3072, -3072, -3072);
		R_FoggedSkyBoxPolyVec ( 3072, -3072, -3072);
		R_FoggedSkyBoxPolyVec ( 3072, -3072,  3072);
	glEnd();

	glBegin(GL_QUADS);
		R_FoggedSkyBoxPolyVec ( 3072, -3072,  3072);
		R_FoggedSkyBoxPolyVec ( 3072,  3072,  3072);
		R_FoggedSkyBoxPolyVec (-3072,  3072,  3072);
		R_FoggedSkyBoxPolyVec (-3072, -3072,  3072);
	glEnd();

	glBegin(GL_QUADS);
		R_FoggedSkyBoxPolyVec ( 3072,  3072, -3072);
		R_FoggedSkyBoxPolyVec ( 3072, -3072, -3072);
		R_FoggedSkyBoxPolyVec (-3072, -3072, -3072);
		R_FoggedSkyBoxPolyVec (-3072,  3072, -3072);
	glEnd();

	glColor3f (1,1,1);
	glEnable  (GL_TEXTURE_2D);
}